import pytest
import asyncio
import aiohttp

from kong.client import Kong, KongError


TESTS = ('test', 'foo', 'pippo')
CONSUMERS = ('an-xxxx-test', 'test-xx', 'test-yy')


@pytest.fixture(autouse=True)
def loop():
    """Return an instance of the event loop."""
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture()
async def cli(loop):
    session = aiohttp.ClientSession(loop=loop)
    async with Kong(session=session) as cli:
        await cleanup(cli)
        yield cli
        await cleanup(cli)


@pytest.fixture()
async def service(cli):
    return await cli.services.create(
        name='test', host='example.upstream', port=8080
    )


@pytest.fixture()
async def consumer(cli, service):
    await service.plugins.create(name='jwt')
    consumer = await cli.consumers.create(username='test-xx')
    return consumer


async def cleanup(cli):
    for name in TESTS:
        try:
            await cli.services.remove(name)
        except KongError as exc:
            if not exc.status == 404:
                raise
    for name in CONSUMERS:
        try:
            await cli.consumers.delete(name)
        except KongError as exc:
            if not exc.status == 404:
                raise



import pytest

from kong.client import KongError


async def test_consumer(cli, consumer):
    assert consumer.username == 'test-xx'
    assert consumer.get('custom_id', '') == ''


async def test_jwt_create(cli, consumer):
    jwt = await consumer.create_jwt()
    assert jwt['consumer_id'] == consumer.id
    data = await consumer.jwts()
    assert data
    data = [d for d in data if d['id'] == jwt['id']]
    assert data


async def test_jwt_delete(cli, consumer):
    jwt = await consumer.create_jwt()
    assert jwt['consumer_id'] == consumer.id
    await consumer.delete_jwt(jwt['id'])
    with pytest.raises(KongError):
        await consumer.delete_jwt(jwt['id'])


async def test_key_auth_create(cli, consumer):
    auth = await consumer.create_key_auth()
    assert auth['consumer_id'] == consumer.id


async def test_key_auth_delete(cli, consumer):
    auth = await consumer.create_key_auth()
    assert auth['consumer_id'] == consumer.id
    await consumer.delete_key_auth(auth['id'])
    with pytest.raises(KongError):
        await consumer.delete_key_auth(auth['id'])


async def test_group(cli, consumer):
    r = await consumer.create_acls('a')
    assert r['consumer_id'] == consumer.id
    assert r['group'] == 'a'
    r = await consumer.create_acls('b')
    assert r['consumer_id'] == consumer.id
    assert r['group'] == 'b'



from .components import CrudComponent, KongEntity, KongError


class Consumers(CrudComponent):

    def wrap(self, data):
        return Consumer(self, data)

    async def apply_json(self, data):
        if not isinstance(data, list):
            data = [data]
        result = []
        for entry in data:
            if not isinstance(entry, dict):
                raise KongError('dictionary required')
            username = entry.pop('username', None)
            groups = entry.pop('groups', [])
            if not username:
                raise KongError('Consumer username is required')
            try:
                consumer = await self.get(username)
            except KongError as exc:
                if exc.status == 404:
                    consumer = await self.create(username=username, **entry)
                else:
                    raise
            else:
                if entry:
                    consumer = await self.update(username, **entry)
            acls = await consumer.acls()
            current_groups = dict(((a['group'], a) for a in acls))
            for group in groups:
                if group not in current_groups:
                    await consumer.create_acls(group)
                else:
                    current_groups.pop(group)

            for acl in current_groups.values():
                await consumer.delete_acls(acl['id'])

            result.append(consumer.data)

        return result


class Consumer(KongEntity):

    @property
    def username(self):
        return self.data.get('username')

    async def jwts(self):
        url = f'{self.url}/jwt'
        result = await self.cli.execute(url, 'GET')
        return result['data']

    def create_jwt(self):
        url = f'{self.url}/jwt'
        return self.cli.execute(
            url, 'POST',
            headers={'Content-Type': 'application/x-www-form-urlencoded'}
        )

    def delete_jwt(self, id):
        url = f'{self.url}/jwt/{id}'
        return self.cli.execute(url, 'DELETE')

    async def key_auths(self):
        url = f'{self.url}/key-auth'
        result = await self.cli.execute(url, 'GET')
        return result['data']

    def create_key_auth(self):
        url = f'{self.url}/key-auth'
        return self.cli.execute(
            url, 'POST',
            headers={'Content-Type': 'application/x-www-form-urlencoded'}
        )

    def delete_key_auth(self, id):
        url = f'{self.url}/key-auth/{id}'
        return self.cli.execute(url, 'DELETE')

    def create_acls(self, group):
        url = f'{self.url}/acls'
        return self.cli.execute(url, 'POST', json=dict(group=group))

    def delete_acls(self, id):
        url = f'{self.url}/acls/{id}'
        return self.cli.execute(url, 'DELETE')

    def acls(self, **params):
        params['consumer_id'] = self.id
        return self.cli.acls.get_list(**params)



from .components import CrudComponent, KongEntity, KongError
from .routes import ServiceRoutes
from .plugins import ServicePlugins
from .utils import local_ip


REMOVE = frozenset(('absent', 'remove'))
LOCAL_HOST = frozenset(('localhost', '127.0.0.1'))


class Services(CrudComponent):
    """Kong API component"""
    def wrap(self, data):
        return Service(self, data)

    async def remove(self, id):
        s = await self.get(id)
        await s.routes.delete_all()
        await self.delete(id)

    async def apply_json(self, data):
        """Apply a JSON data object for a service
        """
        if not isinstance(data, list):
            data = [data]
        result = []
        for entry in data:
            if not isinstance(entry, dict):
                raise KongError('dictionary required')
            ensure = entry.pop('ensure', None)
            name = entry.pop('name', None)
            routes = entry.pop('routes', [])
            plugins = entry.pop('plugins', [])
            host = entry.pop('host', None)
            if host in LOCAL_HOST:
                host = local_ip()
            if not name:
                raise KongError('Service name is required')
            if ensure in REMOVE:
                if await self.has(name):
                    await self.remove(name)
                continue
            # backward compatible with config entry
            config = entry.pop('config', None)
            if isinstance(config, dict):
                entry.update(config)
            if await self.has(name):
                srv = await self.update(name, host=host, **entry)
            else:
                srv = await self.create(name=name, host=host, **entry)
            srv.data['routes'] = await srv.routes.apply_json(routes)
            srv.data['plugins'] = await srv.plugins.apply_json(plugins)
            result.append(srv.data)
        return result


class Service(KongEntity):
    """Object representing a service
    """
    @property
    def plugins(self):
        return ServicePlugins(self, 'plugins')

    @property
    def routes(self):
        return ServiceRoutes(self, 'routes')

    @property
    def name(self):
        return self.data['name']

    @property
    def host(self):
        return self.data.get('host')




import json
import asyncio

import click

import yaml as _yaml

from . import __version__
from .client import Kong, KongError
from .utils import local_ip


@click.command()
@click.option(
    '--version',
    is_flag=True,
    default=False,
    help='Display version and exit'
)
@click.option(
    '--ip',
    is_flag=True,
    default=False,
    help='Show local IP address'
)
@click.option(
    '--key-auth',
    help='Create or display an authentication key for a consumer'
)
@click.option(
    '--yaml', type=click.File('r'),
    help='Yaml configuration to upload'
)
@click.pass_context
def kong(ctx, version, ip, key_auth, yaml):
    if version:
        click.echo(__version__)
    elif ip:
        click.echo(local_ip())
    elif key_auth:
        return _run(_auth_key(ctx, key_auth))
    elif yaml:
        return _run(_yml(ctx, yaml))
    else:
        click.echo(ctx.get_help())


def _run(coro):
    return asyncio.get_event_loop().run_until_complete(coro)


async def _yml(ctx, yaml):
    async with Kong() as cli:
        try:
            result = await cli.apply_json(_yaml.load(yaml))
            click.echo(json.dumps(result, indent=4))
        except KongError as exc:
            raise click.ClickException(str(exc))


async def _auth_key(ctx, consumer):
    async with Kong() as cli:
        try:
            c = await cli.consumers.get(consumer)
            keys = await c.key_auths()
            if keys:
                key = keys[0]
            else:
                key = await c.create_key_auth()
            click.echo(json.dumps(key, indent=4))
        except KongError as exc:
            raise click.ClickException(str(exc))


def main():     # pragma    nocover
    kong()



from itertools import zip_longest

from .components import ServiceEntity, KongEntity
from .utils import as_list
from .plugins import RoutePlugins


class ServiceRoutes(ServiceEntity):
    """Routes associates with a service
    """
    def wrap(self, data):
        return Route(self, data)

    def create(self, skip_error=None, **params):
        params['service'] = dict(id=self.root.id)
        return self.execute(self.url, 'post', json=params,
                            wrap=self.wrap, skip_error=skip_error)

    async def apply_json(self, data):
        if not isinstance(data, list):
            data = [data]
        routes = await self.get_list()
        result = []
        for d, route in zip_longest(data, routes):
            if not d:
                if route:
                    await self.delete(route.id)
                continue
            plugins = d.pop('plugins', [])
            as_list('hosts', d)
            as_list('paths', d)
            as_list('methods', d)
            if not route:
                route = await self.create(**d)
            else:
                route = await self.update(route.id, **d)
            route.data['plugins'] = await route.plugins.apply_json(plugins)
            result.append(route.data)
        return result


class Route(KongEntity):
    """Object representing a route
    """
    @property
    def plugins(self):
        return RoutePlugins(self, 'plugins')



from .components import CrudComponent, KongEntity


class Snis(CrudComponent):
    """Kong SNI API component"""
    def wrap(self, data):
        return Sni(self, data)

    async def apply_json(self, data):
        """Apply a JSON data object for a service
        """
        if not isinstance(data, list):
            data = [data]
        result = []
        for entry in data:
            name = entry.pop('name')
            if await self.has(name):
                sni = await self.update(name, **entry)
            else:
                sni = await self.create(name=name, **entry)
            result.append(sni.data)
        return result


class Sni(KongEntity):
    """Object representing an sni
    """



from .components import CrudComponent, KongEntity


class Acls(CrudComponent):
    """Kong API component"""
    def wrap(self, data):
        return Acl(self, data)


class Acl(KongEntity):
    """Object representing a acl
    """



import os
import pytest
import yaml

from kong.client import KongError


PATH = os.path.dirname(__file__)


def test_client(cli):
    assert cli.session


async def test_create_service(cli):
    c = await cli.services.create(
        name='test', host='example.upstream', port=8080
    )
    assert c.name == 'test'
    assert c.host == 'example.upstream'
    assert c.id


async def test_update_service(cli):
    await cli.services.create(name='test', host='example.upstream', port=8080)
    c = await cli.services.update('test', host='test.upstream')
    assert c.name == 'test'
    assert c.host == 'test.upstream'


async def test_routes(cli):
    await cli.services.create(name='test', host='example.upstream', port=8080)
    c = await cli.services.get('test')
    routes = await c.routes.get_list()
    assert len(routes) == 0
    route = await c.routes.create(hosts=['example.com'])
    assert route['service']['id'] == c.id


async def test_json(cli):
    with open(os.path.join(PATH, 'test.yaml')) as fp:
        manifest = yaml.load(fp)
    await cli.apply_json(manifest)
    srv = await cli.services.get('foo')
    routes = await srv.routes.get_list()
    assert len(routes) == 2
    #
    # check plugins
    plugins = await srv.plugins.get_list()
    assert len(plugins) == 2


async def test_json2(cli):
    with open(os.path.join(PATH, 'test2.yaml')) as fp:
        manifest = yaml.load(fp)
    await cli.apply_json(manifest)
    srv = await cli.services.get('foo')
    routes = await srv.routes.get_list()
    assert len(routes) == 1
    #
    # check plugins
    plugins = await srv.plugins.get_list()
    assert len(plugins) == 2


async def test_add_certificate(cli):
    c = await cli.certificates.create(
        cert='-----BEGIN CERTIFICATE-----...',
        key='-----BEGIN RSA PRIVATE KEY-----...'
    )
    assert c.id
    assert len(c.data['snis']) == 0
    await cli.certificates.delete(c.id)


async def test_hedge_cases(cli):
    with pytest.raises(KongError):
        await cli.apply_json([])

    with pytest.raises(KongError):
        with open(os.path.join(PATH, 'test3.yaml')) as fp:
            await cli.apply_json(yaml.load(fp))

    assert str(cli) == cli.url


async def test_json_plugins(cli):
    with open(os.path.join(PATH, 'test4.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))


async def test_json_route_plugins(cli):
    with open(os.path.join(PATH, 'test6.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))
    with open(os.path.join(PATH, 'test6.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))
    srv = await cli.services.get('pippo')
    plugins = await srv.plugins.get_list()
    assert len(plugins) == 1
    routes = await srv.routes.get_list()
    assert len(routes) == 1
    plugins = await routes[0].plugins.get_list()
    assert len(plugins) == 3
    cs = await cli.consumers.get('an-xxxx-test')
    acls = await cs.acls()
    assert len(acls) == 2

    with open(os.path.join(PATH, 'test61.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))
    srv = await cli.services.get('pippo')
    plugins = await srv.plugins.get_list()
    assert len(plugins) == 0
    routes = await srv.routes.get_list()
    assert len(routes) == 1
    plugins = await routes[0].plugins.get_list()
    assert len(plugins) == 1
    cs = await cli.consumers.get('an-xxxx-test')
    acls = await cs.acls()
    assert len(acls) == 1


async def test_snis(cli):
    c1 = await cli.certificates.create(
        cert='-----BEGIN CERTIFICATE-----...',
        key='-----BEGIN RSA PRIVATE KEY-----...'
    )
    c2 = await cli.certificates.create(
        cert='-----BEGIN CERTIFICATE-----...',
        key='-----BEGIN RSA PRIVATE KEY-----...'
    )
    config = {
        'snis': [
            {
                'name': 'a1.example.com',
                'ssl_certificate_id': c1['id'],
            },
            {
                'name': 'a2.example.com',
                'ssl_certificate_id': c2['id'],
            },
        ]
    }
    resp = await cli.apply_json(config)
    snis = resp['snis']

    # CREATE
    for sni in snis:
        sni.pop('created_at')
    assert snis == config['snis']

    # UPDATE
    config['snis'][0]['ssl_certificate_id'] = c2['id']
    config['snis'][1]['ssl_certificate_id'] = c1['id']
    resp = await cli.apply_json(config)
    snis = resp['snis']

    for sni in snis:
        sni.pop('created_at')
    assert snis == config['snis']

    # GET

    snis = await cli.snis.get_list()
    assert len(snis) == 2
    snis = {
        sni.data['name']: sni.data['ssl_certificate_id']
        for sni in snis
    }
    expected = {
        sni['name']: sni['ssl_certificate_id']
        for sni in config['snis']
    }
    assert snis == expected


async def test_ensure_remove(cli):
    with open(os.path.join(PATH, 'test6.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))
    assert await cli.services.has('pippo') is True
    with open(os.path.join(PATH, 'test7.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))
    assert await cli.services.has('pippo') is False
    with open(os.path.join(PATH, 'test7.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))
    assert await cli.services.has('pippo') is False


async def test_pagination(cli):
    with open(os.path.join(PATH, 'test9.yaml')) as fp:
        await cli.apply_json(yaml.load(fp))
    consumers = []
    async for consumer in cli.consumers.paginate():
        consumers.append(consumer)
    assert len(consumers) >= 2


async def test_paginate_params(cli, consumer):
    await consumer.create_acls('test1')
    await consumer.create_acls('test2')
    consumer2 = await cli.consumers.create(username='an-xxxx-test')
    consumer3 = await cli.consumers.create(username='test-yy')
    await consumer2.create_acls('test2')
    await consumer3.create_acls('test3')
    users = [u async for u in cli.acls.paginate(group='test1', size=1)]
    assert len(users) == 1
    users = [u async for u in cli.acls.paginate(group='test2', size=1)]
    assert len(users) == 2
    acls = await consumer.acls(group='test2')
    assert len(acls) == 1



import socket
from multidict import MultiDict


def as_list(key, data):
    v = data.get(key)
    if isinstance(v, str):
        v = [v]
    elif not isinstance(v, list):
        v = []
    data[key] = v
    return data


def local_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    try:
        return s.getsockname()[0]
    finally:
        s.close()


def as_params(*, params=None, **kwargs):
    params = MultiDict(params if params is not None else {})
    params.update(kwargs)
    return params



import copy
import os
import json

import aiohttp

from .components import KongError, KongResponseError
from .services import Services
from .plugins import Plugins
from .consumers import Consumers
from .certificates import Certificates
from .acls import Acls
from .snis import Snis


__all__ = [
    'Kong',
    'KongError',
    'KongResponseError'
]


class Kong:
    url = os.environ.get('KONG_URL', 'http://127.0.0.1:8001')

    def __init__(self, url: str=None, session: object=None) -> None:
        self.url = url or self.url
        self.session = session
        self.services = Services(self)
        self.plugins = Plugins(self)
        self.consumers = Consumers(self)
        self.certificates = Certificates(self)
        self.acls = Acls(self)
        self.snis = Snis(self)

    def __repr__(self) -> str:
        return self.url
    __str__ = __repr__

    @property
    def cli(self):
        return self

    async def close(self) -> None:
        if self.session:
            await self.session.close()

    async def __aenter__(self) -> object:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    async def execute(self, url, method=None, headers=None,
                      callback=None, wrap=None, timeout=None, skip_error=None,
                      **kw):
        if not self.session:
            self.session = aiohttp.ClientSession()
        method = method or 'GET'
        headers = headers or {}
        headers['Accept'] = 'application/json, text/*; q=0.5'
        response = await self.session.request(
            method, url, headers=headers, **kw
        )
        if callback:
            return await callback(response)
        if response.status == 204:
            return True
        if response.status >= 400:
            try:
                data = await response.json()
            except Exception:
                data = await response.text()
            raise KongResponseError(response, json.dumps(data, indent=4))
        response.raise_for_status()
        data = await response.json()
        return wrap(data) if wrap else data

    async def apply_json(self, config):
        config = copy.deepcopy(config)
        if not isinstance(config, dict):
            raise KongError('Expected a dict got %s' % type(config).__name__)
        result = {}
        for name, data in config.items():
            if not isinstance(data, list):
                data = [data]
            o = getattr(self, name)
            if not o:
                raise KongError('Kong object %s not available' % name)
            result[name] = await o.apply_json(data)
        return result



import json

from click.testing import CliRunner

from kong import __version__
from kong.cli import kong
from kong.utils import local_ip


def test_empty():
    runner = CliRunner()
    result = runner.invoke(kong, [])
    assert result.exit_code == 0
    assert result.output.startswith('Usage: kong [OPTIONS]')


def test_version():
    runner = CliRunner()
    result = runner.invoke(kong, ['--version'])
    assert result.exit_code == 0
    assert result.output.rstrip() == __version__


def test_ip():
    runner = CliRunner()
    result = runner.invoke(kong, ['--ip'])
    assert result.exit_code == 0
    assert result.output.rstrip() == local_ip()


def test_plugins():
    runner = CliRunner()
    result = runner.invoke(kong, ['--yaml', 'tests/test4.yaml'])
    assert result.exit_code == 0


def test_bad_config():
    runner = CliRunner()
    result = runner.invoke(kong, ['--yaml', 'tests/test5.yaml'])
    assert result.exit_code == 1
    assert result.output.strip() == 'Error: Plugin name not specified'


def test_key_auth():
    runner = CliRunner()
    result = runner.invoke(kong, ['--key-auth', 'foo'])
    assert result.exit_code == 1
    result = runner.invoke(kong, ['--yaml', 'tests/test8.yaml'])
    assert result.exit_code == 0
    result = runner.invoke(kong, ['--key-auth', 'an-xxxx-test'])
    assert result.exit_code == 0
    data = json.loads(result.output.rstrip())
    result = runner.invoke(kong, ['--key-auth', 'an-xxxx-test'])
    assert result.exit_code == 0
    data2 = json.loads(result.output.rstrip())
    assert data['key'] == data2['key']



from .components import CrudComponent, KongEntity


class Certificates(CrudComponent):
    """Kong TLS certificate component"""

    def wrap(self, data):
        return Certificate(self, data)


class Certificate(KongEntity):

    def add_sni(self, host):
        pass



import os

from setuptools import setup, find_packages

import kong


def read(name):
    filename = os.path.join(os.path.dirname(__file__), name)
    with open(filename) as fp:
        return fp.read()


def requirements(name):
    install_requires = []
    dependency_links = []

    for line in read(name).split('\n'):
        if line.startswith('-e '):
            link = line[3:].strip()
            if link == '.':
                continue
            dependency_links.append(link)
            line = link.split('=')[1]
        line = line.strip()
        if line:
            install_requires.append(line)

    return install_requires, dependency_links


meta = dict(
    version=kong.__version__,
    description=kong.__doc__,
    name='aio-kong',
    author='Luca Sbardella',
    author_email="luca@quantmind.com",
    maintainer_email="luca@quantmind.com",
    url="https://github.com/lendingblock/aio-kong",
    license="BSD",
    long_description=read("readme.md"),
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,
    zip_safe=False,
    install_requires=requirements('dev/requirements.txt'),
    entry_points={
        "console_scripts": [
            "kong = kong.cli:main"
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: JavaScript',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Utilities'
    ]
)


if __name__ == '__main__':
    setup(**meta)
